/*----------------------------------------------------------------------
  File    : xbcview.c
  Contents: Bayes classifier visualization program for X11
  Author  : Christian Borgelt
  History : 2000.01.05 file created from file xskel.c
            2000.01.12 adapted to module dlgutil
            2000.01.14 naive Bayes classifier loading programmed
            2000.01.17 data table loading programmed
            2000.01.18 colored markers for data points added
            2000.01.27 log file replaced by terminal output
            2000.01.30 adapted to data files without a class column
            2000.02.02 general ellipses drawing function added
            2000.02.03 data format settings made more flexible
            2000.02.08 fuzzy cluster loading programmed
            2000.02.18 adapted to new module nbvload
            2001.02.19 fuzzy membership degree coloring added
            2002.05.07 variable marker size added
            2003.04.12 clustering result visualization added
            2003.05.18 bug in (co)variance handling for clusters fixed
            2004.04.27 redrawing simplified with function bvn_dist
            2005.05.15 drawing of data markers improved
            2007.01.10 adapted to modified bcload.h
            2007.02.14 adapted to modified module xaw/dialog
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/AsciiText.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "xshades.h"
#include "bcload.h"

#if ATTID_MAX > INT_MAX
#error "xbcview requires ATTID_MAX <= INT_MAX"
#endif
#if VALID_MAX > INT_MAX
#error "xbcview requires VALID_MAX <= INT_MAX"
#endif
#if DIMID_MAX > INT_MAX
#error "xbcview requires DIMID_MAX <= INT_MAX"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- extensions --- */
#define MINXEXT      120        /* minimal x-extension of window */
#define MINYEXT       90        /* minimal y-extension of window */
#if 0
#define INITXEXT     320        /* initial x-extension of window */
#define INITYEXT     320        /* initial y-extension of window */
#else
#define INITXEXT     400        /* initial x-extension of window */
#define INITYEXT     400        /* initial y-extension of window */
#endif
#define LISTWD       150        /* minimal list width */
#define LISTHT       120        /* minimal list width */

/* --- marker flags --- */
#define MRK_DATA    0x01        /* mark data points */
#define MRK_COLOR   0x02        /* use color on colored shades */
#define MRK_CENTER  0x10        /* mark centers of normal dists. */
#define MRK_ELL1S   0x20        /* draw 1 sigma ellipses */
#define MRK_ELL2S   0x40        /* draw 2 sigma ellipses */
#define MRK_ELL3S   0x80        /* draw 3 sigma ellipses */

/* --- distribution types --- */
#define DT_PROB        0        /* probabilistic */
#define DT_FUZZY       1        /* fuzzy (probabilistic) */
#define DT_POSS        2        /* fuzzy (possibilistic) */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-9)       /* initialization failed */
#define E_WIDGET    (-10)       /* widget creation failed */
#define E_DIALOG    (-11)       /* cannot create dialog box */
#define E_COLOR     (-12)       /* color allocation failed */
#define E_GRAPH     (-13)       /* cannot create graphics context */
#define E_DLOAD     (-14)       /* loading data failed */

#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Action Function Prototypes
----------------------------------------------------------------------*/
static void ld_class (Widget w, XEvent *e, String *s, Cardinal *c);
static void ld_data  (Widget w, XEvent *e, String *s, Cardinal *c);
static void format   (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit     (Widget w, XEvent *e, String *s, Cardinal *c);
static void colors   (Widget w, XEvent *e, String *s, Cardinal *c);
static void marks    (Widget w, XEvent *e, String *s, Cardinal *c);
static void atts     (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw   (Widget w, XEvent *e, String *s, Cardinal *c);
static void about    (Widget w, XEvent *e, String *s, Cardinal *c);
static void dclick   (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xbcview.rsc"          /* fallback resources */

static XtActionsRec actions[] = {
  { "ld_class", ld_class },     /* File    > Load Classifier... */
  { "ld_data",  ld_data  },     /* File    > Load Data... */
  { "format",   format   },     /* File    > Data Format... */
  { "quit",     quit     },     /* File    > Quit */
  { "redraw",   redraw   },     /* Actions > Redraw */
  { "colors",   colors   },     /* Actions > Colors... */
  { "marks",    marks    },     /* Actions > Markers... */
  { "atts",     atts     },     /* Actions > Attributes... */
  { "about",    about    },     /* Help    > About XBCView... */
  { "dclick",   dclick   },     /* double click in attribute list */
  { "db_focus", db_focus },     /* set focus to input gadget */
  { "db_next",  db_next  },     /* switch to next input gadget */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "XBCView.Error",
                         "no error",                               0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "XBCView.Error",
                         "not enough memory",                      0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "XBCView.Warning",
                         "cannot open file:\n%s",                  1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "XBCView.Warning",
                         "read error on file:\n%s",                1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "XBCView.Warning",
                         "write error on file:\n%s",               1 },
  /* E_PARSE    -5 */  { "warning", "parse",   "XBCView.Warning",
                         "parse error on file:\n%s\n(see terminal "
                         "output for more information)",           1 },
  /* E_CREATE   -6 */  { "warning", "create",  "XBCView.Error",
                         "cannot create classifier",               0 },
  /* E_ATTCNT   -7 */  { "warning", "attcnt",  "XBCView.Error",
                         "no numeric attributes",                  0 },
  /* E_DATA     -8 */  { "warning", "data",    "XBCView.Error",
                         "a classifier must be loaded first",      0 },
  /* E_INIT     -9 */  { "error",   "init",    "XBCView.Error",
                         "initialization failed",                  0 },
  /* E_WIDGET  -10 */  { "error",   "widget",  "XBCView.Error",
                         "widget creation failed",                 0 },
  /* E_DIALOG  -11 */  { "warning", "dialog",  "XBCView.Warning",
                         "cannot create dialog box",               0 },
  /* E_COLOR   -12 */  { "warning", "color",   "XBCView.Warning",
                         "color allocation failed",                0 },
  /* E_GRAPH   -13 */  { "error",   "gc",      "XBCView.Error",
                         "cannot create graphics context",         0 },
  /* E_DLOAD   -14 */  { "warning", "data",    "XBCView.Error",
                         "error loading data:\n%s",                1 },
  /*           -15 */  { "error", "unknown",   "XBCView.Error",
                         "unknown error",                          0 }};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for drawing */

/* --- property sheets --- */
static PSHEET  *psh_fmt   = NULL;  /* data format */
static PSHEET  *psh_color = NULL;  /* color settings */
static PSHEET  *psh_marks = NULL;  /* marker settings */

/* --- dialog box widgets --- */
static Widget  db_atts  = NULL; /* 'Attributes...' dialog box */
static Widget  di_h_att;        /* attribute list for horizontal */
static Widget  di_h_min;        /* input gadgets for minimal */
static Widget  di_h_max;        /* and maximal horizontal value */
static Widget  di_v_att;        /* attribute list for vertical */
static Widget  di_v_min;        /* input gadgets for minimal */
static Widget  di_v_max;        /* and maximal vertical value */

/* --- color and marker information --- */
static PIXEL   black;           /* black, white, and grey */
static PIXEL   white;           /* for markers on colored shades */
static PIXEL   grey;
static SHADES  *shades  = NULL; /* color shades for normal dists. */
static SHADES  *datcols = NULL; /* colors for data points/markers */
static int     mrkflgs  = 0x73; /* flags for markers */
static int     mrksize  = 6;    /* size of data markers */

static struct {                 /* --- color information --- */
  int    dist;                  /* distribution type */
  int    shdbase;               /* shade base */
  int    shdcnt;                /* number of shades */
  int    colcnt;                /* number of colors */
  double coloff;                /* color offset (in degrees) */
  double ndwgt;                 /* weight of normal distribution */
  double cowgt;                 /* complement of this weight */
} colinfo = { DT_PROB, SHD_WHITE, 32, 1, 0.0, 0.8, 0.2 };

/*----------------------------------------------------------------------
  Error Reporting Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (!XtIsTransientShell(widget))
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  int          k;               /* minimal error code */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  k = 1-(int)(sizeof(errmsgs)/sizeof(*errmsgs));
  if (code < k) code = k;       /* check error code against minimum */
  if (code > 0) code = E_DLOAD; /* check for a data load error */
  msg = errmsgs -code;          /* get error message data */
  if (!msg->name) msg = errmsgs -k;
  va_start(va, code);           /* get list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is 'error' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is 'warning' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*----------------------------------------------------------------------
  Color Functions
----------------------------------------------------------------------*/

static void col_delete (void)
{                               /* --- delete color shades */
  if (datcols) { shd_delete(datcols); datcols = NULL; }
  if (shades)  { shd_delete(shades);  shades  = NULL; }
}  /* col_delete() */           /* delete all shades/colors */

/*--------------------------------------------------------------------*/

static int col_create (void)
{                               /* --- create color shades */
  int      clscnt;              /* number of classes/clusters */
  Colormap colmap;              /* default colormap */

  if (!attset) return 0;        /* check for an existing classifier */
  col_delete();                 /* delete existing shades/colors */
  if      (nbc) clscnt = nbc_clscnt(nbc);
  else if (fbc) clscnt = fbc_clscnt(fbc);
  else if (lvq) clscnt = lvq_cnt(lvq);
  else          clscnt = cls_clscnt(clset);
  if (colinfo.colcnt > 0)       /* get the number of classes/clusters */
    colinfo.colcnt = clscnt;    /* and note the number of colors */
  colmap = DefaultColormap(display, screen);
  shades = shd_create(display, colmap, colinfo.coloff,  colinfo.colcnt,
                                       colinfo.shdbase, colinfo.shdcnt);
  if (!shades) return -1;       /* create color shades for distribs. */
  datcols = shd_create(display, colmap, colinfo.coloff, clscnt,
                                        SHD_BLACK,      1);
  if (!datcols) { col_delete(); return -1; }
  return 0;                     /* get colors for data points */
}  /* col_create() */           /* and return 'ok' */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_class (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Classifier...' callback */
  int r;                        /* result of bc_load() */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  r = bc_load(fname);           /* load a classifier/set of clusters */
  if (r) { error(r, fname); return; }
  if (col_create() != 0) {      /* get new color shades */
    error(E_COLOR); return; }   /* and marker colors */
  if ((ptrdiff_t)ok < 2)        /* if not called on startup, */
    redraw(NULL,NULL,NULL,NULL);/* redraw the window contents */
}  /* dcb_class() */

/*--------------------------------------------------------------------*/

static void dcb_data (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Data...' callback */
  int r;                        /* result of bc_data() */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok || !attset) return;   /* if not ok pressed, abort */
  r = bc_data(fname);           /* load a data table */
  if (r) { error(r, (r > 0) ? bc_msg : fname); return; }
  mrkflgs |= MRK_DATA;          /* enforce marks at the data points */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_data() */

/*--------------------------------------------------------------------*/

static void dcb_format (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Data Format...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  datafmt.names = (int)(ptrdiff_t)psh_getval(psh_fmt, "names");
  strncpy(datafmt.recseps, psh_getval(psh_fmt, "record"),  FMTCHRLEN);
  datafmt.recseps[FMTCHRLEN] = '\0';  /* get record separators */
  strncpy(datafmt.fldseps, psh_getval(psh_fmt, "field"),   FMTCHRLEN);
  datafmt.fldseps[FMTCHRLEN] = '\0';  /* get field separators */
  strncpy(datafmt.blanks,  psh_getval(psh_fmt, "blank"),   FMTCHRLEN);
  datafmt.blanks [FMTCHRLEN] = '\0';  /* get blank characters */
  strncpy(datafmt.nullchs, psh_getval(psh_fmt, "null"),    FMTCHRLEN);
  datafmt.nullchs[FMTCHRLEN] = '\0';  /* get unknown value characters */
  strncpy(datafmt.comment, psh_getval(psh_fmt, "comment"), FMTCHRLEN);
  datafmt.comment[FMTCHRLEN] = '\0';  /* get unknown value characters */
}  /* dcb_format() */

/*--------------------------------------------------------------------*/

static void dcb_colors (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Colors...' callback */
  int  clscnt, k;               /* number of classes/clusters, buffer */
  char buf[32];                 /* temporary buffer */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  k = (int)(ptrdiff_t)psh_getval(psh_color, "dist"); /* get the */
  if      (k == 1) colinfo.dist = DT_FUZZY;     /* distribution */
  else if (k == 2) colinfo.dist = DT_POSS;      /* type */
  else             colinfo.dist = DT_PROB;
  if      (fbc)    clscnt = fbc_clscnt(fbc);
  else if (nbc)    clscnt = nbc_clscnt(nbc);
  else if (clset)  clscnt = cls_clscnt(clset);
  else             clscnt = 1;  /* get the number of classes */
  colinfo.colcnt  = (psh_getval(psh_color, "mode") == (void*)0)
                  ? clscnt : 0; /* get shading mode */
  strncpy(buf, psh_getval(psh_color, "coloff"), 32); buf[31] = '\0';
  colinfo.coloff  = atof(buf);  /* get color offset (angle) */
  colinfo.shdbase = (psh_getval(psh_color, "base") == (void*)0)
                  ? SHD_BLACK : SHD_WHITE;  /* get shade base */
  strncpy(buf, psh_getval(psh_color, "shdcnt"), 32); buf[31] = '\0';
  colinfo.shdcnt  = atoi(buf);  /* get number of shades */
  if (colinfo.shdcnt < 1)          colinfo.shdcnt = 1;
  if (colinfo.shdcnt > SHD_MAXCNT) colinfo.shdcnt = SHD_MAXCNT;
  strncpy(buf, psh_getval(psh_color, "ndwgt"),  32); buf[31] = '\0';
  colinfo.ndwgt   = atof(buf);  /* get weight of normal dist. */
  if (colinfo.ndwgt < 0) colinfo.ndwgt  = 0;
  if (colinfo.ndwgt > 1) colinfo.ndwgt  = 1;
  colinfo.cowgt = 1.0 -colinfo.ndwgt;
  if (col_create() != 0)        /* get new color shades */
    error(E_COLOR);             /* and marker colors */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_colors() */

/*--------------------------------------------------------------------*/

static void dcb_marks (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Markers...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  if (psh_getval(psh_marks, "color"))  mrkflgs |=  MRK_COLOR;
  else                                 mrkflgs &= ~MRK_COLOR;
  if (psh_getval(psh_marks, "center")) mrkflgs |=  MRK_CENTER;
  else                                 mrkflgs &= ~MRK_CENTER;
  if (psh_getval(psh_marks, "ell1s"))  mrkflgs |=  MRK_ELL1S;
  else                                 mrkflgs &= ~MRK_ELL1S;
  if (psh_getval(psh_marks, "ell2s"))  mrkflgs |=  MRK_ELL2S;
  else                                 mrkflgs &= ~MRK_ELL2S;
  if (psh_getval(psh_marks, "ell3s"))  mrkflgs |=  MRK_ELL3S;
  else                                 mrkflgs &= ~MRK_ELL3S;
  if (psh_getval(psh_marks, "data"))   mrkflgs |=  MRK_DATA;
  else                                 mrkflgs &= ~MRK_DATA;
  mrksize = atoi(psh_getval(psh_marks, "size"));
  if (mrksize <= 0) mrksize = 1;/* get toggle states, set flags, */
  redraw(NULL,NULL,NULL,NULL);  /* and redraw the window contents */
}  /* dcb_marks() */

/*--------------------------------------------------------------------*/

static void dcb_atts (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Attributes...' callback */
  int    i;                     /* list index */
  String s;                     /* text in input gadgets */

  if (((ptrdiff_t)ok != 2)      /* if not to keep dialog open */
  && (((ptrdiff_t)ok != 1) || ((ptrdiff_t)call != 2))) {
    XtPopdown(db_atts);         /* close dialog box */
    mn_enable(menu, MN_MENU, 1);/* enable menu bar */
  }
  if ((ptrdiff_t)ok == 0) return;  /* if not ok pressed, abort */

  i = XawListShowCurrent(di_h_att)->list_index;
  if (i != XAW_LIST_NONE) {     /* if an attribute is selected, */
    selatt.h_att = i;           /* note it for the horizontal dir. */
    XtVaGetValues(di_h_min, XtNstring, &s, NULL);
    ranges[selatt.h_att].min = selatt.h_rng.min = atof(s);
    XtVaGetValues(di_h_max, XtNstring, &s, NULL);
    ranges[selatt.h_att].max = selatt.h_rng.max = atof(s);
  }                             /* get the range of values */
  i = XawListShowCurrent(di_v_att)->list_index;
  if (i != XAW_LIST_NONE) {     /* if an attribute is selected, */
    selatt.v_att = i;          /* note it for the vertical dir. */
    XtVaGetValues(di_v_min, XtNstring, &s, NULL);
    ranges[selatt.v_att].min = selatt.v_rng.min = atof(s);
    XtVaGetValues(di_v_max, XtNstring, &s, NULL);
    ranges[selatt.v_att].max = selatt.v_rng.max = atof(s);
  }                             /* get the range of values */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_atts() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_class  (Widget widget, XtPointer client, XtPointer call)
{ ld_class(NULL, NULL, NULL, NULL); }

static void mcb_data   (Widget widget, XtPointer client, XtPointer call)
{ ld_data(NULL, NULL, NULL, NULL); }

static void mcb_format (Widget widget, XtPointer client, XtPointer call)
{ format(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_colors (Widget widget, XtPointer client, XtPointer call)
{ colors(NULL, NULL, NULL, NULL); }

static void mcb_marks  (Widget widget, XtPointer client, XtPointer call)
{ marks(NULL, NULL, NULL, NULL); }

static void mcb_atts   (Widget widget, XtPointer client, XtPointer call)
{ atts(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  List Callback Function
----------------------------------------------------------------------*/

static void lcb_att (Widget widget, XtPointer client, XtPointer call)
{                               /* --- attribute list callback */
  int  i;                       /* list index */
  char buf[32];                 /* conversion buffer */

  i = ((XawListReturnStruct*)call)->list_index;
  if ((i == XAW_LIST_NONE) || !ranges)
    return;                     /* check selected list item */
  widget = ((ptrdiff_t)client == 0) ? di_h_min : di_v_min;
  sprintf(buf, "%g", ranges[i].min);
  XtVaSetValues(widget, XtNstring, buf, NULL);
  widget = ((ptrdiff_t)client == 0) ? di_h_max : di_v_max;
  sprintf(buf, "%g", ranges[i].max);
  XtVaSetValues(widget, XtNstring, buf, NULL);
}  /* lcb_att() */              /* set new range of values */

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void ld_class (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Classifier... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*ld_class"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_class) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* ld_class() */             /* show a file selector box */

/*--------------------------------------------------------------------*/

static void ld_data (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Data... */
  const char *t;                /* buffer for dialog box title */

  if (!attset) {                /* if no classifier has been loaded, */
    error(E_DATA); return; }    /* no data file can be loaded */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*ld_data"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_data) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* ld_data() */              /* show a file selector box */

/*--------------------------------------------------------------------*/

static void format (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Data Format... */
  if (!psh_fmt) {               /* if property sheet not created yet */
    psh_fmt = psh_create("fmtDlg", w_top, 1);
    if (!psh_fmt) return;       /* create a property sheet */
    if ((psh_addline(psh_fmt, "lnames")               != 0)
    ||  (psh_additem(psh_fmt, "names",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_fmt, "names",   "no")        != 0)
    ||  (psh_menuadd(psh_fmt, "names",   "yes")       != 0)
    ||  (psh_addline(psh_fmt, "lrecord")              != 0)
    ||  (psh_additem(psh_fmt, "record",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lfield")               != 0)
    ||  (psh_additem(psh_fmt, "field",   PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lblank")               != 0)
    ||  (psh_additem(psh_fmt, "blank",   PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lnull")                != 0)
    ||  (psh_additem(psh_fmt, "null",    PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lcomment")             != 0)
    ||  (psh_additem(psh_fmt, "comment", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_fmt); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_fmt, "names",   (char*)(ptrdiff_t)datafmt.names);
  psh_setval(psh_fmt, "record",  datafmt.recseps);
  psh_setval(psh_fmt, "field",   datafmt.fldseps);
  psh_setval(psh_fmt, "blank",   datafmt.blanks);
  psh_setval(psh_fmt, "null",    datafmt.nullchs);
  psh_setval(psh_fmt, "comment", datafmt.comment);
  psh_handle(psh_fmt, dcb_format); /* fill input gadgets with current */
}  /* format() */                  /* values and handle dialog */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  bc_clean();                   /* clean up the variables and */
  col_delete();                 /* delete all colors and shades */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate the program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void dclick (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- double click in attribute list */
  dcb_atts(db_atts, (void*)2, NULL);
}  /* dclick() */               /* call dialog box callback for 'ok' */

/*--------------------------------------------------------------------*/

static int listbox (Widget form, Widget left, const char *name)
{                               /* --- create a list box */
  Cardinal n;                   /* widget argument counter */
  Arg      args[16];            /* widget argument list */
  Widget   view, list;          /* viewport and list widget */
  Widget   min, max;            /* minimum and maximum widget */
  Widget   label;               /* buffer for label widgets */
  char     buf[16];             /* buffer for widget name */

  /* --- create list box --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainTop);              n++;
  XtSetArg(args[n], XtNbottom,      XtChainTop);              n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s", name);  /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create title widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNfromVert,    label);                   n++;
  XtSetArg(args[n], XtNallowHoriz,  False);                   n++;
  XtSetArg(args[n], XtNallowVert,   True);                    n++;
  XtSetArg(args[n], XtNuseRight,    True);                    n++;
  XtSetArg(args[n], XtNtop,         XtChainTop);              n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  sprintf(buf, "v%s", name);  /* create the widget name */
  view = XtCreateManagedWidget(buf, viewportWidgetClass,
                               form, args, n);
  if (!view) return -1;       /* create list viewport */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNstring,         "");                   n++;
  XtSetArg(args[n], XtNborderWidth,    0);                    n++;
  XtSetArg(args[n], XtNverticalList,   True);                 n++;
  XtSetArg(args[n], XtNdefaultColumns, 1);                    n++;
  XtSetArg(args[n], XtNforceColumns,   True);                 n++;
  sprintf(buf, "%s_att", name);  /* create the widget name */
  list = XtCreateManagedWidget(buf, listWidgetClass, view, args, n);
  if (!list) return -1;       /* create list widget */

  /* --- create input field for minimum --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  XtSetArg(args[n], XtNfromVert,    view);                    n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s_min", name); /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create min label widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  XtSetArg(args[n], XtNeditType,    XawtextEdit);             n++;
  XtSetArg(args[n], XtNfromHoriz,   label);                   n++;
  XtSetArg(args[n], XtNfromVert,    view);                    n++;
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  min = XtCreateManagedWidget(buf+1, asciiTextWidgetClass,
                                 form, args, n);
  if (!min) return -1;        /* create minimum widget */

  /* --- create input field for maximum --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  XtSetArg(args[n], XtNfromVert,    min);                     n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s_max", name); /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create h_max label widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  XtSetArg(args[n], XtNeditType,    XawtextEdit);             n++;
  XtSetArg(args[n], XtNfromHoriz,   label);                   n++;
  XtSetArg(args[n], XtNfromVert,    min);                     n++;
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  max = XtCreateManagedWidget(buf+1, asciiTextWidgetClass,
                              form, args, n);
  if (!max) return -1;          /* create maximum widget */
  return 0;                     /* return 'ok' */
}  /* listbox() */

/*--------------------------------------------------------------------*/

static void atts (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Attributes... */
  static String empty = "";     /* an empty string */
  XtTranslations tla, tlo;      /* translations for text widgets */
  Widget    shell = NULL, form; /* widgets for dialog shell and form */
  Widget    lview, rview;       /* left and right viewport widget */
  Widget    lh_min, lh_max;     /* label widgets for horizontal */
  Widget    lv_min, lv_max;     /* and vertical minimum/maximum */
  Widget    ok, cancel;         /* widgets for ok and cancel buttons */
  Arg       args[16];           /* widget argument list */
  Cardinal  n;                  /* widget argument counter */
  Position  x,  y;              /* position of created dialog */
  Dimension wd, ht;             /* width and height of created dialog */
  Dimension lblwd, t;           /* label width and temp. buffer */
  int       dd;                 /* default distance of dialog items */

  if (!db_atts) {               /* if dialog box is not created */
    shell = XtCreatePopupShell("attDlg", transientShellWidgetClass,
              w_top, NULL, 0);  /* create a dialog box shell */
    if (!shell) { error(E_DIALOG); return; }
    XtSetArg(args[0], XtNborderWidth, 0);
    form = XtCreateManagedWidget("form", formWidgetClass,
             shell, args, 1);   /* create a form for the box layout */
    if (!form) { XtDestroyWidget(shell); error(E_DIALOG); return; }

    /* --- create listbox for horizontal axis --- */
    if (listbox(form, NULL,  "h") != 0) {
      XtDestroyWidget(shell); error(E_DIALOG); return; }
    lview    = XtNameToWidget(form, "*vh");
    di_h_att = XtNameToWidget(form, "*h_att");
    di_h_min = XtNameToWidget(form, "*h_min");
    di_h_max = XtNameToWidget(form, "*h_max");
    XtAddCallback(di_h_att, XtNcallback, lcb_att, (void*)0);

    /* --- create listbox for vertical axis --- */
    if (listbox(form, lview, "v") != 0) {
      XtDestroyWidget(shell); error(E_DIALOG); return; }
    rview    = XtNameToWidget(form, "*vv");
    di_v_att = XtNameToWidget(form, "*v_att");
    di_v_min = XtNameToWidget(form, "*v_min");
    di_v_max = XtNameToWidget(form, "*v_max");
    XtAddCallback(di_v_att, XtNcallback, lcb_att, (void*)1);

    /* --- create ok button --- */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromVert,    di_h_max);      n++;
    XtSetArg(args[n], XtNleft,        XtChainLeft);   n++;
    XtSetArg(args[n], XtNtop,         XtChainBottom); n++;
    XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
    ok = XtCreateManagedWidget("ok", commandWidgetClass,
           form, args, n);      /* create 'ok' button */
    if (!ok) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    XtAddCallback(ok, XtNcallback, dcb_atts, (void*)1);

    /* --- create cancel button --- */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromHoriz,   ok);            n++;
    XtSetArg(args[n], XtNfromVert,    di_v_max);      n++;
    XtSetArg(args[n], XtNtop,         XtChainBottom); n++;
    XtSetArg(args[n], XtNright,       XtChainRight);  n++;
    XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
    cancel = XtCreateManagedWidget("cancel", commandWidgetClass,
               form, args, n);  /* create 'cancel' button */
    if (!cancel) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    XtAddCallback(cancel, XtNcallback, dcb_atts, (void*)0);

    /* --- adapt widths of labels and gadgets --- */
    XtSetArg(args[0], XtNwidth, &t); lblwd = 1;
    lh_min = XtNameToWidget(form, "*lh_min");
    XtGetValues(lh_min, args, 1); if (t > lblwd) lblwd = t;
    lh_max = XtNameToWidget(form, "*lh_max");
    XtGetValues(lh_max, args, 1); if (t > lblwd) lblwd = t;
    lv_min = XtNameToWidget(form, "*lv_min");
    XtGetValues(lv_min, args, 1); if (t > lblwd) lblwd = t;
    lv_max = XtNameToWidget(form, "*lv_max");
    XtGetValues(lv_max, args, 1); if (t > lblwd) lblwd = t;
    XtVaSetValues(lh_min,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lh_max,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lv_min,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lv_max,   XtNwidth, lblwd, NULL);
    XtVaGetValues(di_h_min, XtNwidth, &wd, NULL);
    #ifdef LINUX                /* if Linux Athena Widgets */
    XtVaGetValues(form,     XtNdefaultDistance, &dd, NULL);
    #else                       /* get the default distance, */
    dd = 4;                     /* if Sun or HP Athena Widgets, */
    #endif                      /* set the default distance */
    wd = (Dimension)(wd +lblwd +dd); /* compute listbox width */
    if (wd < LISTWD) wd = LISTWD;               /* and set it */
    XtVaSetValues(rview,    XtNwidth, wd, NULL);
    XtVaSetValues(lview,    XtNwidth, wd, NULL);
    XtVaSetValues(ok,       XtNwidth, wd, NULL);
    XtVaSetValues(cancel,   XtNwidth, wd, NULL);
    wd = (Dimension)(wd -lblwd -dd); /* compute input field width */
    XtVaSetValues(di_h_min, XtNwidth, wd, NULL);
    XtVaSetValues(di_h_max, XtNwidth, wd, NULL);
    XtVaSetValues(di_v_min, XtNwidth, wd, NULL);
    XtVaSetValues(di_v_max, XtNwidth, wd, NULL);

    /* --- install translations --- */
    tla = XtParseTranslationTable(
            "<Btn1Down>,<Btn1Up>: Set()Notify()\n"
            "<Btn1Down>(2+):      dclick()");
    XtAugmentTranslations(di_h_att, tla);
    XtAugmentTranslations(di_v_att, tla);
    tla = XtParseTranslationTable(
            "<Key>Home  : beginning-of-line()\n"
            "<Key>End   : end-of-line()\n");
    tlo = XtParseTranslationTable(
            "Shift<Key>Tab   : db_next(-1)\n"
            "<Key>Tab        : db_next(+1)\n"
            "Shift<Key>Return: db_close(ok)\n"
            "<Key>Return     : db_close(apply)\n"
            "<Key>Escape     : db_close(cancel)\n"
            "!Alt<Key>o      : db_close(ok)\n"
            "!Alt<Key>c      : db_close(cancel)\n"
            "<Btn1Down>      : db_focus()\n");
    XtAugmentTranslations (di_h_min, tla);
    XtOverrideTranslations(di_h_min, tlo);
    XtAugmentTranslations (di_h_max, tla);
    XtOverrideTranslations(di_h_max, tlo);
    XtAugmentTranslations (di_v_min, tla);
    XtOverrideTranslations(di_v_min, tlo);
    XtAugmentTranslations (di_v_max, tla);
    XtOverrideTranslations(di_v_max, tlo);

    XtInstallAllAccelerators(form, form);
    db_atts = shell;            /* install actions and accelerators */
  }                             /* and note the created dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */

  /* --- fill dialog box fields --- */
  if (!names || (attcnt <= 0)){ /* if the name list is not valid */
    XawListChange(di_h_att, (const char**)&empty, 1, 0, 1);
    XawListChange(di_v_att, (const char**)&empty, 1, 0, 1);
    XtSetArg(args[0], XtNstring, "");
    XtSetValues(di_h_min, args, 1);  /* clear the list boxes */
    XtSetValues(di_h_max, args, 1);  /* and the input gadgets */
    XtSetValues(di_v_min, args, 1);
    XtSetValues(di_v_max, args, 1); }
  else {                        /* if the name list is valid */
    XawListChange(di_h_att, (const char**)names, attcnt, 0, 1);
    XawListHighlight(di_h_att, selatt.h_att);
    XawListChange(di_v_att, (const char**)names, attcnt, 0, 1);
    XawListHighlight(di_v_att, selatt.v_att);
    sprintf(buf, "%g", selatt.h_rng.min);
    XtVaSetValues(di_h_min, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.h_rng.max);
    XtVaSetValues(di_h_max, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.v_rng.min);
    XtVaSetValues(di_v_min, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.v_rng.max);
    XtVaSetValues(di_v_max, XtNstring, buf, NULL);
  }                             /* set attribute names and ranges */
  XtVaSetValues(XtParent(di_h_att), XtNheight, LISTHT, NULL);
  XtVaSetValues(XtParent(di_v_att), XtNheight, LISTHT, NULL);

  /* --- show dialog box --- */
  XtVaGetValues(w_top,   XtNx, &x, XtNy, &y, NULL);
  XtVaSetValues(db_atts, XtNx,  x, XtNy,  y, NULL);
  XtPopup(db_atts, XtGrabExclusive);   /* place and show dialog box */
  XtSetKeyboardFocus(db_atts, di_h_min);  /* and set keyboard focus */

  if (shell) {                  /* if dialog box has been created */
    XtVaGetValues(shell, XtNwidth,   &wd, XtNheight,   &ht, NULL);
    XtVaSetValues(shell, XtNminWidth, wd, XtNminHeight, ht, NULL);
    XtOverrideTranslations(shell,
      XtParseTranslationTable("<Message>WM_PROTOCOLS: db_close()"));
    XSetWMProtocols(XtDisplay(shell), XtWindow(shell),
      &wm_delwin, 1);           /* override window manager delete */
  }
}  /* atts() */

/*--------------------------------------------------------------------*/

static void colors (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Colors... */
  void *p;                      /* temporary buffer */
  char buf[32];                 /* ditto, for texts */

  if (!psh_color) {             /* if property sheet not created yet */
    psh_color = psh_create("colorDlg", w_top, 1);
    if (!psh_color) return;     /* create a property sheet */
    if ((psh_addline(psh_color, "ldist")               != 0)
    ||  (psh_additem(psh_color, "dist",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_color, "dist",   "prob")      != 0)
    ||  (psh_menuadd(psh_color, "dist",   "fuzzy")     != 0)
    ||  (psh_menuadd(psh_color, "dist",   "poss")      != 0)
    ||  (psh_addline(psh_color, "lmode")               != 0)
    ||  (psh_additem(psh_color, "mode",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_color, "mode",   "color")     != 0)
    ||  (psh_menuadd(psh_color, "mode",   "grey")      != 0)
    ||  (psh_addline(psh_color, "lcoloff")             != 0)
    ||  (psh_additem(psh_color, "coloff", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_color, "lbase")               != 0)
    ||  (psh_additem(psh_color, "base",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_color, "base",   "black")     != 0)
    ||  (psh_menuadd(psh_color, "base",   "white")     != 0)
    ||  (psh_addline(psh_color, "lshdcnt")             != 0)
    ||  (psh_additem(psh_color, "shdcnt", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_color, "lndwgt")              != 0)
    ||  (psh_additem(psh_color, "ndwgt",  PSH_EDIT, 0) != 0)) {
      psh_delete(psh_color); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if      (colinfo.dist == DT_FUZZY) p = (void*)1;
  else if (colinfo.dist == DT_POSS)  p = (void*)2;
  else                               p = (void*)0;
  psh_setval(psh_color, "dist", p);     /* set distribution type */
  psh_setval(psh_color, "mode",         /* set shade mode */
    (const char*)(ptrdiff_t)((colinfo.colcnt > 0) ? 0 : 1));
  sprintf(buf, "%g", colinfo.coloff);
  psh_setval(psh_color, "coloff", buf); /* set color offset */
  psh_setval(psh_color, "base",         /* set shade base */
    (const char*)(ptrdiff_t)((colinfo.shdbase == SHD_BLACK) ? 0 : 1));
  sprintf(buf, "%d", colinfo.shdcnt);
  psh_setval(psh_color, "shdcnt", buf); /* set number of shades */
  sprintf(buf, "%g", colinfo.ndwgt);
  psh_setval(psh_color, "ndwgt", buf);  /* set weight of normal dist. */
  psh_handle(psh_color, dcb_colors);    /* and handle dialog */
}  /* colors() */

/*--------------------------------------------------------------------*/

static void marks (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Markers... */
  char buf[32];                 /* buffer for texts */

  if (!psh_marks) {             /* if property sheet not created yet */
    psh_marks = psh_create("marksDlg", w_top, 1);
    if (!psh_marks) return;     /* create a property sheet */
    if ((psh_addline(psh_marks, NULL)                    != 0)
    ||  (psh_additem(psh_marks, "color",  PSH_TOGGLE, 0) != 0)
    ||  (psh_addline(psh_marks, NULL)                    != 0)
    ||  (psh_additem(psh_marks, "center", PSH_TOGGLE, 0) != 0)
    ||  (psh_addline(psh_marks, NULL)                    != 0)
    ||  (psh_additem(psh_marks, "ell1s",  PSH_TOGGLE, 0) != 0)
    ||  (psh_addline(psh_marks, NULL)                    != 0)
    ||  (psh_additem(psh_marks, "ell2s",  PSH_TOGGLE, 0) != 0)
    ||  (psh_addline(psh_marks, NULL)                    != 0)
    ||  (psh_additem(psh_marks, "ell3s",  PSH_TOGGLE, 0) != 0)
    ||  (psh_addline(psh_marks, NULL)                    != 0)
    ||  (psh_additem(psh_marks, "data",   PSH_TOGGLE, 0) != 0)
    ||  (psh_addline(psh_marks, "lsize")                 != 0)
    ||  (psh_additem(psh_marks, "size",   PSH_EDIT,   0) != 0)) {
      psh_delete(psh_marks); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_marks, "color",
            (void*)(ptrdiff_t)(mrkflgs & MRK_COLOR  ? 1 : 0));
  psh_setval(psh_marks, "center",
            (void*)(ptrdiff_t)(mrkflgs & MRK_CENTER ? 1 : 0));
  psh_setval(psh_marks, "ell1s",
            (void*)(ptrdiff_t)(mrkflgs & MRK_ELL1S  ? 1 : 0));
  psh_setval(psh_marks, "ell2s",
            (void*)(ptrdiff_t)(mrkflgs & MRK_ELL2S  ? 1 : 0));
  psh_setval(psh_marks, "ell3s",
            (void*)(ptrdiff_t)(mrkflgs & MRK_ELL3S  ? 1 : 0));
  psh_setval(psh_marks, "data",
            (void*)(ptrdiff_t)(mrkflgs & MRK_DATA   ? 1 : 0));
  sprintf(buf, "%d", mrksize); psh_setval(psh_marks, "size", buf);
  psh_handle(psh_marks,dcb_marks); /* set marker flags */
}  /* marks() */                   /* and handle dialog */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  static Dimension owd = 0;     /* old window width */
  static Dimension oht = 0;     /* old window height */
  int        i, k, n;           /* loop variables, buffer */
  int        clscnt;            /* number of classes/clusters */
  int        colcnt;            /* number of colors */
  Dimension  wd, ht;            /* window extensions */
  XRectangle rc;                /* clipping rectangle */
  int        x, y, x_prev;      /* pixel coordinates */
  double     rx, ry;            /* real  coordinates */
  double     dx, dy;            /* standard deviations */
  double     xoff, yoff;        /* offsets of window borders */
  double     xscl, yscl;        /* scaling factors */
  double     val, sum, max;     /* function value, sum and maximum */
  PIXEL      pxl, pxl_prev;     /* current and previous pixel value */
  PIXEL      mrk;               /* pixel value for markers */
  int        col, shd, shdcnt;  /* color, shade, and number of shades */
  XEvent     de;                /* dummy event */
  double     a, b;              /* transformation constants */
  int        xid, yid, clsid;   /* attribute identifiers */
  int        xtype, ytype;      /* attribute types */
  TPLID      r;                 /* loop variable for tuples */
  TUPLE      *tpl;              /* to traverse the data tuples */
  float      f;                 /* temporary buffer */
  int        cmin, cmax;        /* minimal and maximal coordinate */
  double     t0, t1, t2;        /* temporary buffers */
  char       *msg;              /* buffer for message */

  /* --- initialize graphics --- */
  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  if (!bvnorm || !attset || !shades) {
    XClearWindow(display, window);
    msg = "Load a classifier or a cluster set.";
    XDrawString(display, window, gc, 8, 16, msg, (int)strlen(msg));
    return;                     /* if there is no drawing data */
  }                             /* clear window and draw message */
  XtVaGetValues(w_view, XtNwidth, &wd, XtNheight, &ht, NULL);
  if (e && (e->type == Expose)  /* if the window was exposed, */
  &&  (wd == owd) && (ht == oht)) {        /* but not resized */
    rc.x      = (short)e->xexpose.x;       /* build and set */
    rc.y      = (short)e->xexpose.y;       /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted); }
  else {                        /* if the window was resized */
    rc.x = rc.y = 0;            /* or this is a forced redraw, */
    rc.width  = wd;             /* redraw all of the window */
    rc.height = ht;             /* (set full window rectangle) */
    while (XCheckTypedWindowEvent(display, window, Expose, &de));
  }                             /* remove all exposure events */
  owd = wd; oht = ht;           /* note (new) window dimensions */

  /* --- initialize normal distributions --- */
  clscnt = bc_bvnorm();         /* init. the normal distributions */
  if (clscnt < 0) return;       /* and check for success */

  /* --- draw rectangle contents --- */
  xoff = ranges[selatt.h_att].min;  /* get coordinate offsets and */
  yoff = ranges[selatt.v_att].max;  /* compute the scaling factors */
  xscl = (ranges[selatt.h_att].max -xoff) /(wd -1);
  yscl = (yoff -ranges[selatt.v_att].min) /(ht -1);
  a    = 1.0/clscnt;            /* compute transformation constants */
  b    = (clscnt > 1) ? clscnt/(clscnt-1.0) : 1;
  colcnt = shd_colcnt(shades);  /* get the number of colors */
  shdcnt = shd_shdcnt(shades);  /* and the number of shades */
  pxl_prev = shd_pixel(shades, 0, 0);
  for (y = rc.y; y < rc.y +rc.height; y++) {
    for (x = x_prev = rc.x +rc.width; --x >= rc.x; ) {
      rx = xoff +x *xscl;       /* traverse the pixels in the rect. */
      ry = yoff -y *yscl;       /* and compute their real coordinates */
      max = sum = 0; col = 0;   /* traverse the classes/clusters */
      if (colinfo.dist >= DT_FUZZY) {  /* if fuzzy memberships */
        for (i = clscnt; --i >= 0; ) {
          val  = bvn_dist(bvnorm +i, rx, ry);
          sum += val = (val > 0) ? 1/val : 1e12;
          if (val > max) { max = val; col = i; }
        }                       /* compute Mahalanobis distance */
        if (colinfo.dist == DT_POSS) { /* if possibilistic coloring */
          val = max /(1 +max);} /* compute relative intensity */
        else {                  /* if probabilistic fuzzy coloring */
          val = (sum > 0) ? (colinfo.ndwgt *(max/sum)
                          +  colinfo.cowgt *(max/sum -a) *b) : 0;
        } }                     /* compute relative intensity */
      else {                    /* if probabilistic coloring */
        for (i = clscnt; --i >= 0; ) {
          sum += val = bvn_eval(bvnorm +i, rx, ry);
          if (val > max) { max = val; col = i; }
        }                       /* compute class probabilities */
        val = (sum > 0) ? (colinfo.ndwgt *(max/bvn_max(bvnorm +col))
                         + colinfo.cowgt *(max/sum -a) *b) : 0;
      }                         /* compute the relative intensity */
      shd = (int)(shdcnt *val); /* and from it the color shade */
      if (shd    >= shdcnt) shd = shdcnt-1;
      if (shd    <  0)      shd = 0;
      if (colcnt <= 0)      col = 0;
      pxl = shd_pixel(shades, col, shd);
      if (pxl == pxl_prev) continue;
      XSetForeground(display, gc, pxl_prev);
      XDrawLine(display, window, gc, x_prev, y, x, y);
      x_prev   = x;             /* draw a line segment and note */
      pxl_prev = pxl;           /* the end of the line segment */
    }                           /* and the new pixel value */
    XSetForeground(display, gc, pxl_prev);
    XDrawLine(display, window, gc, x_prev, y, x, y);
  }                             /* draw last line segment */

  /* --- draw markers --- */
  mrk = ((colcnt > 0) && (shd_shdbase(shades) == SHD_BLACK))
      ? white : black;          /* get the default marker color */
  if (mrkflgs & (MRK_CENTER|MRK_ELL1S|MRK_ELL2S|MRK_ELL3S)) {
    if (colcnt > 0)             /* black/white markers on color */
      XSetForeground(display, gc, mrk);
    for (i = clscnt; --i >= 0; ) {  /* traverse the classes */
      if (colcnt <= 0)          /* on grey use the class color */
        XSetForeground(display, gc, shd_pixel(datcols, i, 0));
      rx = bvn_ex(bvnorm +i);   /* get the expected values */
      ry = bvn_ey(bvnorm +i);   /* (coordinates of the center) */
      if (mrkflgs & MRK_CENTER){/* if to mark the centers */
        x = (int)((rx -xoff) /xscl +0.5) -1;
        y = (int)((yoff -ry) /yscl +0.5) -1;
        XDrawRectangle(display, window, gc, x, y, 2, 2);
      }                         /* draw a 3x3 pixel rectangle */
      dx = bvn_dx(bvnorm +i);   /* get the standard deviations */
      dy = bvn_dy(bvnorm +i);   /* (extensions of the ellipse) */
      for (k = 1; k <= 3; k++){ /* traverse the ellipses */
        if (!(mrkflgs & (MRK_ELL1S << (k-1))))
          continue;             /* skip unselected ellipses */
        cmin = (int)((rx -k*dx -xoff) /xscl +0.5);
        cmax = (int)((rx +k*dx -xoff) /xscl +0.5);
        for (x = cmin; x <= cmax; x++) {
          t0 = xoff +x *xscl;
          bvn_ellx2y(bvnorm +i, k, t0, &t1, &t2);
          y = (int)((yoff -t1) /yscl +0.5);
          XDrawPoint(display, window, gc, x, y);
          y = (int)((yoff -t2) /yscl +0.5);
          XDrawPoint(display, window, gc, x, y);
        }
        cmin = (int)((yoff -ry -k*dy) /yscl +0.5);
        cmax = (int)((yoff -ry +k*dy) /yscl +0.5);
        for (y = cmin; y <= cmax; y++) {
          t0 = yoff -y *yscl;
          bvn_elly2x(bvnorm +i, k, t0, &t1, &t2);
          x = (int)((t1 -xoff) /xscl +0.5);
          XDrawPoint(display, window, gc, x, y);
          x = (int)((t2 -xoff) /xscl +0.5);
          XDrawPoint(display, window, gc, x, y);
        }                       /* for each k sigma ellipse: */
      }                         /* draw the ellipse twice, */
    }                           /* once traversing the x-coordinates */
  }  /* if (mrkflgs .. */       /* and once the y-coordinates */

  /* --- draw data --- */
  if (table                     /* if a data table has been loaded */
  &&  (mrkflgs & MRK_DATA)) {   /* and the marker flag is set */
    if      (nbc) clsid = nbc_trgid(nbc);
    else if (fbc) clsid = fbc_trgid(fbc);
    else          clsid = -1;   /* get the class attribute id. */
    xid   = map_fwd[selatt.h_att]; /* get other attribute ids. */
    yid   = map_fwd[selatt.v_att]; /* and attribute types */
    xtype = att_type(as_att(attset, xid));
    ytype = att_type(as_att(attset, yid));
    pxl   = grey;               /* set default marker color */
    for (r = tab_tplcnt(table); --r >= 0; ) {
      tpl = tab_tpl(table, r);  /* traverse the tuples */
      col = (clsid >= 0) ? tpl_colval(tpl, clsid)->i : NV_NOM;
      if (xtype == AT_INT) {    /* if x-coordinate is integer valued */
        i = tpl_colval(tpl, xid)->i;
        if (isnull(i)) continue;
        rx = (double)i; }       /* get and check the attribute value */
      else {                    /* if x-coordinate is real valued */
        f = tpl_colval(tpl, xid)->f;
        if (isnan(f))  continue;
        rx = (double)f;         /* get and check the attribute value */
      }
      if (ytype == AT_INT) {    /* if y-coordinate is integer valued */
        i = tpl_colval(tpl, yid)->i;
        if (isnull(i)) continue;
        ry = (double)i; }       /* get and check the attribute value */
      else {                    /* if y-coordinate is real valued */
        f = tpl_colval(tpl, yid)->f;
        if (isnan(f))  continue;
        ry = (double)f;         /* get and check the attribute value */
      }                         /* then compute the screen coords. */
      n = mrksize; i = n >> 1;  /* compute the marker size parameters */
      x = (int)((rx -xoff) /xscl +0.5) -i;
      y = (int)((yoff -ry) /yscl +0.5) -i;
      if ((mrkflgs & MRK_COLOR) /* if to mark in color */
      ||  (colcnt <= 0)) {      /* or the shades are grey */
        pxl = ((col < 0) || (col >= shd_colcnt(datcols)))
            ? grey : shd_pixel(datcols, col, 0);
      }                         /* set the class/marker color */
      XSetForeground(display, gc, pxl);
      XFillArc(display, window, gc, x, y, (unsigned)n, (unsigned)n,
               0, 360*64);
      XSetForeground(display, gc, mrk);
      XDrawArc(display, window, gc, x, y, (unsigned)n, (unsigned)n,
               0, 360*64);
    }                           /* draw a circle at the data point */
  }  /* if (table ... */

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg      args[8];             /* widget arguments */
  Cardinal n;                   /* argument counter */
  Colormap colmap;              /* color map of display */
  XColor   xc[2];               /* buffer for colors */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "XBCView",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "ld_class", 0, mcb_class,  0) != 0)
  ||  (mn_additem (menu, "ld_data",  0, mcb_data,   0) != 0)
  ||  (mn_additem (menu, "format",   0, mcb_format, 0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "quit",     0, mcb_quit,   0) != 0)
  ||  (mn_addtitle(menu, "actions",  0, 0,          0) != 0)
  ||  (mn_additem (menu, "atts",     0, mcb_atts,   0) != 0)
  ||  (mn_additem (menu, "colors",   0, mcb_colors, 0) != 0)
  ||  (mn_additem (menu, "marks",    0, mcb_marks,  0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "redraw",   0, mcb_redraw, 0) != 0)
  ||  (mn_addtitle(menu, "help",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "about",    0, mcb_about,  0) != 0)
  ||  (mn_addtitle(menu, NULL,       0, 0,          0) != 0))
    error(E_WIDGET);
  mn_resize(menu, INITXEXT);    /* set width of menu bar */

  /* --- create viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);      n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);       n++;
  XtSetArg(args[n], XtNright,     XtChainRight);     n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom);    n++;
  XtSetArg(args[n], XtNresizable, True);             n++;
  w_view = XtCreateManagedWidget("view",
             coreWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */
  XtVaSetValues(w_view, XtNwidth,     INITXEXT,
                        XtNheight,    INITYEXT, NULL);
  XtVaSetValues(w_top,  XtNminWidth,  MINXEXT,
                        XtNminHeight, MINYEXT,  NULL);
                                /* set window size */
  XtOverrideTranslations(w_view,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_view);   /* and drawable for window */

  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  colmap = DefaultColormap(display, screen);
  if (XAllocNamedColor(display, colmap, "grey", xc+1, xc) == 0)
    error(E_COLOR);             /* allocate colors */
  black  = BlackPixel(display, screen);
  white  = WhitePixel(display, screen);
  grey   = xc[1].pixel;         /* enter the color values */

  /* --- evaluate command line arguments --- */
  if (argc > 1) dcb_class(NULL, (XtPointer)2, (XtPointer)argv[1]);
  if (argc > 2) dcb_data (NULL, (XtPointer)1, (XtPointer)argv[2]);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
